"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qpy/__init__.py $
$Id: __init__.py 28581 2006-08-14 22:11:40Z dbinger $
"""
__version__ = "1.3"

# external code should import these from here.
from qpy.c8 import u8, h8, stringify, html_escape_string
[u8, h8, stringify, html_escape_string] # for qpcheck.py.


